/*! ********************************************
 * smf.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_smf_h
#define INCLUDE_smf_h

#ifndef INCLUDE_GrayBox_h
#include "GrayBox.hpp"
#endif

namespace sora {
namespace graybox {
namespace smf {
/***********************************************
	Class	SMFTrackEvent
************************************************/
class SMFTrackEvent
{
public:
	SMFTrackEvent* next = nullptr;
	int32 deltaTime = 0;

	int32 channel  = 0;
	int32 type     = 0;
	int32 value0   = 0;
	int32 value1   = 0;

	int32 dataSize = 0;
	uint8* data = nullptr;
	
	void clear();
// Constructor
	SMFTrackEvent() = default;
	SMFTrackEvent(int32 deltaT, int32 type, int32 ch, int32 val0 = 0, int32 val1 = 0);
// Copy Constructor
	SMFTrackEvent(const SMFTrackEvent&);
	SMFTrackEvent& operator = (const SMFTrackEvent&);
// Move Constructor
	SMFTrackEvent(SMFTrackEvent&&)noexcept;
	SMFTrackEvent& operator = (SMFTrackEvent&&)noexcept;
// Destructor
	~SMFTrackEvent();
};
/***********************************************
	Class	SMFSequence
************************************************/
class SMFSequence
{
// Instance -----
private:
	bool _created = false;
	int32 _format = 0;
	bool _multiPacketSysEx = true;
	SMFTrackEvent* _track = nullptr;  // Event List - "MTrk" Events
	// Arbitrary Loop
	SMFTrackEvent* _loopStart = nullptr;
	SMFTrackEvent* _loopEnd   = nullptr;
public:
	int32 division = 480;  // "MThd" divison
// getter, setter ---
	void setTrack(SMFTrackEvent* trk, bool move = false);
	SMFTrackEvent* getTrack() { return _track; }
	bool getLoopable(){ return _loopStart != nullptr; }
	SMFTrackEvent* getLoopStart(){ return _loopStart; }
	SMFTrackEvent* getLoopEnd(){ return _loopEnd; }
	bool setLoopPoint(SMFTrackEvent* start = nullptr, SMFTrackEvent* end = nullptr);
	void clearData();
// Constructor
	SMFSequence() = default;
	SMFSequence(uint8* dat, uint32 dataSize, bool multiPacketSysEx = true);
// Copy Constructor
	SMFSequence(const SMFSequence&) = delete;
	SMFSequence& operator = (const SMFSequence&) = delete;
// Move Constructor
	SMFSequence(SMFSequence&&)noexcept;
	SMFSequence& operator = (SMFSequence&&)noexcept;
// Destructor
	~SMFSequence();

	bool loadBytes(uint8* bytes, uint32 dataSize, bool multiPacketSysEx = true);
private:
	bool _read_MTrk(uint8* bytes, uint32 chunkDataSize, SMFTrackEvent* trk);
	void _insertTrackEvent(SMFTrackEvent* &crnt, SMFTrackEvent* ev);
	void _removeUnecEvents(SMFTrackEvent* trk);
	int32 _readVarLeng(uint8* &dat);
};
/***********************************************
	Class	SMFSequencer
************************************************/
class SMFSequencer : public IGrayBoxModule
{
// EventBuffer -----
private:
	MidiEvent* _eventBuffer = nullptr;
	int32 _eventBufferSize = 512;
	int32 _usedEvents = 0;
public:
	void clearMIDIEventBuffer();
	int32 getEventBufferSize() { return _eventBufferSize; }
	int32 setEventBufferSize(int32 size);
// Sequence -----
private:
	SMFSequence* _sequence = nullptr;
	SMFTrackEvent* _track  = nullptr; // Event List - "MTrk" Events
	int32 _division = 480;  // "MThd" divison
	SMFTrackEvent* _loopStart = nullptr;
	SMFTrackEvent* _loopEnd   = nullptr;
public:
	SMFSequence* getSequence() { return _sequence; }
	bool setSequence(SMFSequence* seq);
// Sequencer -----
	int32 midiOutPorts = 0xFFFF;
private:
	/*
	   if (event.type <= _eventFilter) --> process
	   0xE0 = ChannelEvent Only
	   0xFF = include SysEx
	   0xFFFF = include SysEx and MetaEvent
	*/
	int32 _eventFilter = 0xFF;
	float _sampleRate = 44100.0f;
	double _tempo = 500000.0;  // SetTempo (MetaEvent FF51)
	bool _isPlay = false;
	bool _loopPlay = false;
	SMFTrackEvent* _trackPos = nullptr;
	// event delay timer
	static constexpr int32 _EV_TIMER_FPBITS  = 24;
	static constexpr int32 _EV_TIMER_FPVALUE = 1 << _EV_TIMER_FPBITS;
	static constexpr int32 _EV_TIMER_FPMASK  = _EV_TIMER_FPVALUE - 1;
	static constexpr double _EV_TIMER_FPVALUE_N  = double(_EV_TIMER_FPVALUE);
	double _refTime    = 0.0;  // val = _timeDivision(Note Ticks) * (1000000uSec / SampleRate)
	int32 _evTimer     = 0;
	int32 _evTimerAFP  = 0;
	int32 _evTimerStep = 0;
// Sequencer status ---
public:
	float getSampleRate() { return _sampleRate; }
	void setSampleRate(float val);
	bool getOutputMetaMessage() { return _eventFilter == 0xFFFF; }
	void setOutputMetaMessage(bool flag) { _eventFilter = (flag)? 0xFFFF : 0xFF; }
	bool playable() { return _track != nullptr; }
	bool isPlay() { return _isPlay; }
	bool isLoopPlay() { return _loopPlay; }
	bool endOfTrack() { return _trackPos == nullptr; }
// Sequencer Contorols ---
	void reset();
	void initialize();
	double getBPM() { return 60000000.0 / _tempo; }
	void setBPM(double val);
// Sequence Contorol ---
	bool play(bool resume = false);
	void stop(bool pause = false);
	bool loopPlay(bool flag);
	void gotoHead();
// Constructor
	SMFSequencer(int32 eventBufferSize = 512);
// Copy Constructor
	SMFSequencer(const SMFSequencer&) = delete;
	SMFSequencer& operator = (const SMFSequencer&) = delete;
// Move Constructor
	SMFSequencer(SMFSequencer&&)noexcept;
	SMFSequencer& operator = (SMFSequencer&&)noexcept;
// Destructor
	~SMFSequencer();
// Interface - IfGrayBoxModule ---
	void updateSampleRate(float) override;
	void updateAudioChannels(int32) override {}
	void updateProcessBlockSize(int32) override {}
	
	void resetModule() override;
	void initiModule() override;
	void connectModule(GrayBox*) override;
	void disconnectModule() override;
	void updateBypassStatus(bool) override;
	
	void startProcess() override;
	void stopProcess() override;
	void suspendProcess() override;
	void resumeProcess() override;
	void process(ProcessData*) override;
};
}}}
#endif